<?php

namespace Drupal\monarch_entity_yml\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;

/**
 * A Drush commandfile.
 *
 * In addition to this file, you need a drush.services.yml
 * in root of your module, and a composer.json file that provides the name
 * of the services file to use.
 *
 * See these files for an example of injecting Drupal services:
 *   - http://cgit.drupalcode.org/devel/tree/src/Commands/DevelCommands.php
 *   - http://cgit.drupalcode.org/devel/tree/drush.services.yml
 */
class EntityYmlCommands extends DrushCommands {

  /**
   * All references.
   *
   * @var array
   */
  protected $references = [];

  /**
   * All entities exported.
   *
   * @var array
   */
  protected $entities = [];

  /**
   * Export data.
   *
   * @var array
   */
  protected $data = [];

  /**
   * Referenced allowed to be exported.
   *
   * @var array
   */
  protected $exportReferenceTypes = [];

  /**
   * Include non-banned references.
   *
   * @var bool
   */
  protected $includeReferences = FALSE;

  /**
   * A reference of fields that shouldn't be exported.
   *
   * @var array
   */
  protected const BANNED_FIELDS = [
    'node' => [
      'path' => TRUE,
      'revision_timestamp' => TRUE,
      'revision_uid' => TRUE,
      'revision_log' => TRUE,
      'revision_default' => TRUE,
      'revision_translation_affected' => TRUE,
    ],
  ];

  /**
   * Add ancestor entities.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to attempt loading ancestors for.
   * @param bool $export_if_neutral
   *   Export the entity by default unless specifically blacklisted.
   */
  protected function queueEntity(ContentEntityInterface $entity, bool $export_if_neutral = FALSE) {
    if (isset($this->entities[$entity->getEntityTypeId() . ':' . $entity->id()])) {
      return;
    }

    $this->entities[$entity->getEntityTypeId() . ':' . $entity->id()] = TRUE;

    if (method_exists($entity, 'getParentId')) {
      $parentId = $entity->getParentId();

      if (!empty($parentId)) {
        list($entityTypeId, $uuid) = explode(':', $parentId);

        $parentEntity = \Drupal::service('entity.repository')->loadEntityByUuid($entityTypeId, $uuid);

        if (!empty($parentEntity)) {
          $this->queueEntity($parentEntity, $this->includeReferences);
        }
      }
    }

    if (!method_exists($entity, 'getFields')) {
      return;
    }

    $entity_data = [
      'storage' => $entity->getEntityTypeId(),
    ];

    foreach ($entity->getFields() as $name => $field) {
      if (static::BANNED_FIELDS['any'][$name] ?? FALSE || static::BANNED_FIELDS[$entity->getEntityTypeId()][$name] ?? FALSE) {
        continue;
      }

      $entity_data['values'][$name] = $field->getValue();

      if (method_exists($field, 'referencedEntities')) {
        foreach ($field->referencedEntities() as $childEntity) {
          $childKey = $childEntity->getEntityTypeId() . ':' . $childEntity->id();

          if (!($childEntity instanceof ConfigEntityInterface)) {
            $this->references[$childKey] = $childKey;
          }

          if ($childEntity instanceof ContentEntityInterface) {
            if ($export_if_neutral && !isset($this->exportReferenceTypes[$childEntity->getEntityTypeId()]) || $this->exportReferenceTypes[$childEntity->getEntityTypeId()] ?? NULL) {
              $this->queueEntity($childEntity, $this->includeReferences);
            }
          }
        }
      }
    }

    if ($entity instanceof ContentEntityInterface) {
      if ($export_if_neutral && !isset($this->exportReferenceTypes[$entity->getEntityTypeId()]) || $this->exportReferenceTypes[$entity->getEntityTypeId()] ?? NULL) {
        array_push($this->data, $entity_data);
      }
    }
  }

  /**
   * Export entity as configuration.
   *
   * @param string $entity_list
   *   Entity ids.
   * @param array $options
   *   The options provided by drush.
   *
   * @options with-references Include all references (unless explicitly excluded).
   * @usage entity_yml:dump entity_type:id > file.yml
   *   Export an entity to yml.
   * @usage entity_yml:dump entity_type:id,id,id... > file.yml
   *   Export multiple entities of a type to yml.
   * @usage entity_yml:dump entity_type:* > file.yml
   *   Export all entities of a type to yml.
   * @usage entity_yml:dump entity_type:+ > file.yml
   *   Export all secondary references to entities of a type to yml.
   * @usage entity_yml:dump entity_type:- > file.yml
   *   Exclude entity type from export (but still check it for references, if
   *   included by other criteria).
   * @usage entity_yml:dump ...;...;... > file.yml
   *   Specify multiple criteria.
   * @usage entity_yml:dump "... ... ..." > file.yml
   *   Specify multiple criteria.
   * @usage entity_yml:dump "entity_type:id another_entity_type:* another_entity_type:+" > file.yml
   *   Export multiple entities and their referenced entities to yml.
   *
   * @command entity_yml:dump
   * @aliases eyd
   */
  public function entityExport(string $entity_list = NULL, array $options = ['with-references' => FALSE]) {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $this->includeReferences = $options['with-references'] ? TRUE : FALSE;
    $this->data = [];
    $this->references = [];
    $this->entities = [];
    $this->exportReferenceTypes = [];

    preg_match_all('/([^\s;:]+):([^\s;:]+)/i', $entity_list, $matches, PREG_SET_ORDER);

    foreach ($matches as $match) {
      if ($match[2] === '*' || $match[2] === '+') {
        $this->exportReferenceTypes[$match[1]] = $this->exportReferenceTypes[$match[1]] ?? TRUE;
      }
      elseif ($match[2] === '-') {
        $this->exportReferenceTypes[$match[1]] = FALSE;
      }
    }

    foreach ($matches as $match) {
      try {
        $storage = $entityTypeManager->getStorage($match[1]);

        if (empty($storage)) {
          $this->stderr()->writeln('Entity storage not found: ' . $match[1]);
          continue;
        }

        if ($match[2] === '*') {
          $res = $storage->loadMultiple();
        }
        elseif ($match[2] === '+') {
          continue;
        }
        else {
          $res = $storage->loadMultiple(explode(',', $match[2]));
        }

        foreach ($res as $entity) {
          if (!($entity instanceof ContentEntityInterface)) {
            $this->stderr()->writeln('Skipping non-content entity: ' . $entity->getEntityTypeId() . ':' . $entity->id());
            continue;
          }

          if (empty($entity)) {
            $this->stderr()->writeln('Entity not found: ' . $match[0]);
            continue;
          }

          $this->queueEntity($entity, TRUE);
        }
      }
      catch (\Throwable $e) {
        $this->stderr()->writeln($e->getMessage());
      }
    }

    $this->output()->writeln(Yaml::encode($this->data));

    if (!empty($this->references)) {
      $this->references = array_filter(array_values($this->references), function ($key) {
        return empty($this->entities[$key]);
      });

      $this->stderr()->writeln("Unexported References: " . implode(' ', $this->references));
    }
  }

  /**
   * Export entity as configuration.
   *
   * @usage entity_yml:import < file.yml
   *  Imports yml and attempts to recreate entities defined within.
   *
   * @command entity_yml:import
   * @aliases eyi
   */
  public function entityImport() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $this->data = stream_get_contents(STDIN);
    $this->data = Yaml::decode($this->data);

    foreach ($this->data as $definition) {
      if (isset($definition['storage']) && is_string($definition['storage']) && isset($definition['values']) && is_array($definition['values'])) {
        try {
          $entityTypeManager->getStorage($definition['storage'])->create($definition['values'])->save();
        }
        catch (\Throwable $e) {
          $this->stderr()->writeln("*******************************************************************************");
          $this->stderr()->writeln("Error importing: " . print_r($definition, TRUE) . "\n" . $e->getMessage());
          $this->stderr()->writeln("*******************************************************************************");
          exit(1);
        }
      }
    }
  }

  /**
   * List entity types eligible for export.
   *
   * @usage entity_yml:list
   *  List entities eligible for export.
   *
   * @command entity_yml:list
   * @aliases eyl
   */
  public function entityList() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $entity_definitions = $entityTypeManager->getDefinitions();

    $list = [];

    foreach ($entity_definitions as $definition) {
      if ($definition instanceof ContentEntityTypeInterface) {
        $list[] = $definition->id();
      }
    }

    if (empty($list)) {
      $this->stderr()->writeln('No entity types eligible for export.');
    }
    else {
      $this->output()->writeln('Entity types eligible for export:');

      sort($list);

      foreach ($list as $entity_type) {
        $this->output()->writeln("\t" . $entity_type);
      }
    }
  }

}
